import numpy as np
import matplotlib.pyplot as plt
import astropy
import archive_explorer

class HabZoneEvaluator:
    """Class for Habitable Zone Calculations
    
    Contains functions to calculate whether the planet is in the habitable zone, per several different models of the habitable zone.

    """
    def __init__(self):
        pass

    def conservative_habzone(self, hostname=None, t_eff=None, dec=None, period=None, mandr=None):
        """Conservative Habitable Zone

        Evaluation of whether a given planet is in the conservative habitable zone, as given by the runaway greenhouse and maximum greenhouse limits in Kopparapu et al. 2013

        Args:
            hostname (tuple): Names of the host stars you are interested in (as strings with single quotations)
            t_eff (tuple): Tuple of minimum and maximum stellar effective temperatures you are interested in
            dec (tuple): Tuple of minimum and maximum declinations you are interested in
            period (tuple): Tuple of minimum and maximum planet orbital periods you are interested in (in days)
            mandr (Boolean, default=False): Require that there be measured mass and radius values

        Returns:
            DataFrame: Table of the planet systems and properties produced by ArchiveExplorer.query_exo(), with additional columns for the inner and outer radius of the conservative habitable zone and whether the planet is in the conservative habitable zone.
        """
        #Inner radius = runaway greenhouse (per Kopparapu et al. 2013)
        #Outer radius = maximum greenhouse
        sol_flux_in = 1.0512
        a_in = 1.3242e-4
        b_in = 1.5418e-8
        c_in = -7.9895e-12
        d_in = -1.8328e-15

        sol_flux_out = 0.3438
        a_out = 5.8942e-5
        b_out = 1.6558e-9
        c_out = -3.0045e-12
        d_out = -5.2983e-16
        explorer = archive_explorer.ArchiveExplorer()
        data = explorer.query_exo(hostname=hostname, t_eff=t_eff, dec=dec, period=period, mandr=mandr)
        inner_rads = []
        outer_rads = []
        in_hz = []
        for index, row in data.iterrows():
            semimajor = explorer._orb_dist(row)
            t_star = float(row['st_teff']) - 5780
            pl_stflux = float((10**row['st_lum'])/semimajor**2) #Stellar luminosity is in units of log(L/L_sun)
            cons_inner_stflux = float((sol_flux_in + a_in*t_star + b_in*(t_star**2) + c_in*(t_star**3) + d_in*(t_star**4))/np.sqrt(1 - float(row['pl_orbeccen'])**2))
            cons_outer_stflux = float((sol_flux_out + a_out*t_star + b_out*(t_star**2) + c_out*(t_star**3) + d_out*(t_star**4))/np.sqrt(1 - float(row['pl_orbeccen'])**2))
            cons_inner_rad = np.sqrt(float((10**row['st_lum'])/cons_inner_stflux))
            cons_outer_rad = np.sqrt(float((10**row['st_lum'])/cons_outer_stflux))
            inner_rads.append(cons_inner_rad)
            outer_rads.append(cons_outer_rad)
            if pl_stflux > cons_outer_stflux and pl_stflux < cons_inner_stflux:
                in_hz.append(True)
            else:
                in_hz.append(False)
        data['Conservative Inner Radius (AU)'] = inner_rads
        data['Conservative Outer Radius (AU)'] = outer_rads
        data['In Conservative Habitable Zone'] = in_hz
        return data

    def optimistic_habzone(self, hostname=None, t_eff=None, dec=None, period=None, mandr=None):
        """Optimistic Habitable Zone

        Evaluation of whether a given planet is in the optimistic habitable zone, as given by the recent Venus and early Mars limits in Kopparapu et al. 2013

        Args:
            hostname (tuple): Names of the host stars you are interested in (as strings)
            t_eff (tuple): Tuple of minimum and maximum stellar effective temperatures you are interested in
            dec (tuple): Tuple of minimum and maximum declinations you are interested in
            period (tuple): Tuple of minimum and maximum planet orbital periods you are interested in (in days)
            mandr (Boolean, default=False): Require that there be measured mass and radius values

        Returns:
            DataFrame: Table of the planet systems and properties produced by ArchiveExplorer.query_exo(), with additional columns for the inner and outer radius of the conservative habitable zone and whether the planet is in the optimistic habitable zone.
        """
        #Inner radius = recent Venus (per Kopparapu et al. 2013)
        #Outer radius = early Mars
        sol_flux_in = 1.7753
        a_in = 1.4316e-4
        b_in = 2.9875e-9
        c_in = -7.5702e-12
        d_in = -1.1635e-15

        sol_flux_out = 0.3179
        a_out = 5.4513e-5
        b_out = 1.5313e-9
        c_out = -2.7786e-12
        d_out = -4.8997e-16
        explorer = archive_explorer.ArchiveExplorer()
        data = explorer.query_exo(hostname=hostname, t_eff=t_eff, dec=dec, period=period, mandr=mandr)
        inner_rads = []
        outer_rads = []
        in_hz = []
        for index, row in data.iterrows():
            semimajor = explorer._orb_dist(row)
            t_star = float(row['st_teff'])- 5780
            pl_stflux = float((10**row['st_lum'])/semimajor**2) #Stellar luminosity is in units of log(L/L_sun)
            opt_inner_stflux = float((sol_flux_in + a_in*t_star + b_in*(t_star**2) + c_in*(t_star**3) + d_in*(t_star**4))/np.sqrt(1 - float(row['pl_orbeccen'])**2))
            opt_outer_stflux = float((sol_flux_out + a_out*t_star + b_out*(t_star**2) + c_out*(t_star**3) + d_out*(t_star**4))/np.sqrt(1 - float(row['pl_orbeccen'])**2))
            opt_inner_rad = np.sqrt(float((10**row['st_lum'])/opt_inner_stflux))
            opt_outer_rad = np.sqrt(float((10**row['st_lum'])/opt_outer_stflux))
            inner_rads.append(opt_inner_rad)
            outer_rads.append(opt_outer_rad)
            if pl_stflux > opt_outer_stflux and pl_stflux < opt_inner_stflux:
                in_hz.append(True)
            else:
                in_hz.append(False)
        data['Optimistic Inner Radius (AU)'] = inner_rads
        data['Optimistic Outer Radius (AU)'] = outer_rads
        data['In Optimistic Habitable Zone'] = in_hz
        return data


import numpy as np
import pandas as pd
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive

class ArchiveExplorer:

    """ ArchiveExplorer class - Handles queries between the user and NASA's Exoplanet Archive
    """

    cols = ['gaia_id', 'ra', 'dec', 'pl_pubdate', 'pl_name', 'hostname', 'st_mass', 'st_teff', 'st_lum',
             'pl_orbper','pl_orbsmax', 'pl_masse', 'pl_msinie','pl_rade', 'pl_eqt','pl_orbeccen', 'pl_dens']
    G = 6.743e-11 # m^3 kg^-1 s^-2
    m_earth = 5.9722e24 # mass of earth in kg
    m_sun = 1.989e30 # mass of sun in kg
    au = 1.496e11 # 1 AU in m
    day = 60 * 60 * 24 # day in seconds

    def __init__(self):
        pass

    def _classify_planet_by_density(self, density_ratio):
        """ Classifies a planet's type based on its density

        Args:
            density_ratio: A pandas series or numpy array of density ratios (float)
        
        Returns:
            String classificatoin of density (Gas planet, water, world, or Rocky planet)
        """
#         if np.isnan(density_ration):
#           return None
        if density_ratio < 0.4:
            return "Gas"
        elif density_ratio < 0.7:
            return "Water"
        else:
            return "Rocky"
    

    def query_exo(self, table='pscomppars', hostname=None, t_eff=None, dec=None, 
                 period=None, mandr=False, cols=None):
        """ Queries the NASA Exoplanet archive

        Calculates orbital distance and planet density and adds them to query results

        Args:
            table (optional): Table to pull from (typically 'ps' or 'pscomppars')
            hostname (optional): Specify hostname of a star or set of stars (e.g. 'Kepler-%')
            t_eff (optional): Range of effective temperatures [lo, hi]
            dec (optional): Declination range to search [lo, hi]
            period (optional): Period range to search [lo, hi]
            mandr (optional): Specifies that both mass and radius must be non-null
            cols: (optional): List of additional column names as string
            Default columns:
                gaia_id: Gaia ID of the star
                ra: Right Ascension (star)
                dec: Declination (star)
                pl_pubdate: Initial date of publication of the planet's data
                pl_name: Most commonly used planet name
                hostname: Name of host star
                st_mass: Mass of host star (in solar masses)
                st_teff: Effective temperature of host star (in Kelvin)
                st_lum: Log luminosity of host star (in log10(Solar))
                pl_orbper: Orbital period of planet
                pl_orbsmax: Orbital distance of planet
                pl_masse: Mass of planet (in Earth masses)
                pl_msinie: Minimum mass (in Earth masses)
                pl_rade: Radius of planet (in Earth radii)
                pl_eqt: Equilibrium temperature of the planet (in Kelvin)
                pl_orbeccen: Orbital eccentricity of the planet
                pl_dens: Density of planet (in g/cm^3)
        
        Returns:
            results: Results of query as a pandas dataframe.
            Orbital distance will be a new column 'pl_orbdist' (in AU), 
            and planet density classification will be in 'pl_type' (string)
        """
        
        # Add default cuts, unless user specified
        _range = lambda param, minmax: f"{param}>{minmax[0]} and {param}<{minmax[1]}"

        # Cut on eccentricity (important for the equations)
        cuts = ["pl_orbeccen<0.3"]
        if cols is not None: [self.cols.append(col) for col in cols if col not in self.cols]

        # Other cuts
        if mandr: cuts.append("pl_masse is not null and pl_rade is not null")
        if hostname is not None: cuts.append(f"hostname like '{hostname}'")
        if t_eff is not None: cuts.append(_range('st_teff', t_eff))
        if dec is not None: cuts.append(_range('dec', dec))
        if period is not None: cuts.append(_range('pl_orbper', period))

        # Query exoplanet archive
        tab = NasaExoplanetArchive.query_criteria(table=table, 
                                                  select=', '.join(self.cols),
                                                  where=' and '.join(cuts)
                                                  ).to_pandas()
        
        
        # Calculate orbital distance and add to table
        tab['pl_orbdist'] = self._orb_dist(tab)
        tab['pl_type'] = tab['pl_dens'].apply(lambda x: self._classify_planet_by_density(x) 
                                                  if pd.notnull(x) else None)

        # Drop duplicates (last first) if the table includes them
        if table!='pscomppars':
            tab.sort_values(by='pl_pubdate', ascending=False, ignore_index=True, inplace=True)
            tab.drop_duplicates(subset=['gaia_id', 'pl_name'], keep='first', inplace=True, ignore_index=True)

        self.results = tab
        return tab
    
    def _orb_dist(self, data):
        """ Calculates orbital distance from orbital period 
        Args:
            data: A pandas dataframe obtained from query_expo
        Returns:
            A pandas series of orbital distance in AU
        """
        r = np.cbrt((self.G * data.st_mass * self.m_sun / (4 * np.pi**2)) 
                    * (data.pl_orbper * self.day)**2)
        return r / self.au # orbital distance in AU



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.offsetbox import (AnchoredOffsetbox, AuxTransformBox,
                                  DrawingArea, TextArea, VPacker)
import astropy
from matplotlib.patches import Circle, Ellipse, Annulus
import archive_explorer
import hab_zone


class PlotHZ:
    """Class: PlotHZ
    Useful Plots to visualize planets in the habitable zone

    """
    def __init__(self):
        pass
         
    
    def plot_hab(self, hostname=None, pl_name=None,sma=None, eccen=None, cons_in=None, cons_out=None, opt_in=None, opt_out=None):

        """Plot Habitable Zone
        Visual representation of the planet orbit and Habitable Zone around the star
        The conservative and optimistic habitable zone are plotted in concentric circles,
        while the planet orbit will be an ellipse depending on eccentricity

        Args:
            hostname (string)
            sma (float): Semi-major axis in AU
            eccen (float): eccentricity
            cons_in: Inner bound of conservative habitable zone in AU
            cons_out: Outer bound of conservative habitable zone in AU
            opt_in: Inner bound of optimistic habitable zone in AU
            opt_out: Outer bound of optimistic habitable zone in AU
        
        Returns:
            pyplot
        """
        exp = archive_explorer.ArchiveExplorer()
        if hostname:
            tab = exp.query_exo(hostname=hostname)
        else:
            tab = exp.query_exo()

        if tab.empty:
            print("No matching exoplanet data found.")
            return

        if pl_name:
            planet_row = tab[tab["pl_name"] == pl_name]
            if planet_row.empty:
                print(f"No planet named {pl_name} found for {hostname}.")
                return
            tab = planet_row
        else:
            tab = tab.iloc[[0]]  # default to first planet

        eccen = tab["pl_orbeccen"].iloc[0]
        sma = tab["pl_orbsmax"].iloc[0]

        eval = hab_zone.HabZoneEvaluator()

        cons_data = eval.conservative_habzone(hostname=hostname)
        opt_data = eval.optimistic_habzone(hostname=hostname)

        if cons_data.empty or opt_data.empty:
            print(f"Habitable zone data not found for {hostname}.")
            return

        cons_in = cons_data['Conservative Inner Radius (AU)'].iloc[0]
        cons_out = cons_data['Conservative Outer Radius (AU)'].iloc[0]

        opt_in = opt_data['Optimistic Inner Radius (AU)'].iloc[0]
        opt_out = opt_data['Optimistic Outer Radius (AU)'].iloc[0]

        cons_zone = Annulus((0, 0), cons_out, cons_out - cons_in, color='green', alpha=0.8, label="Conservative HZ")
        opt_zone = Annulus((0, 0), opt_out, opt_out - opt_in, color='green', alpha=0.4, label="Optimistic HZ")

        # Orbital parameters
        a = sma  # semi-major axis
        b = np.sqrt(1 - eccen**2) * a  # semi-minor axis
        focus_offset = eccen * a  # distance from center to star (focus)
        orbit = Ellipse((-focus_offset, 0), 2 * a, 2 * b, color='black', fill=False, label="Planet Orbit")


        fig, ax = plt.subplots()

        ax.add_patch(cons_zone)
        ax.add_patch(opt_zone)

        ax.add_patch(orbit)

        ax.set_xlabel("Distance (AU)")
        ax.set_ylabel("Distance (AU)")

        ax.set_aspect('equal')

        ax.plot(0, 0, marker='*', markersize=10, color='gold', zorder=5)

        max_radius = max(cons_out, opt_out, sma * (1 + eccen))  # add margin for orbit
        ax.set_xlim(-1.2*max_radius, 1.2*max_radius)
        ax.set_ylim(-1.2*max_radius, 1.2*max_radius)
        ax.legend()

        plt.title(f"Habitable Zone and Orbit for {pl_name}")
        plt.show()

        return

    def plot_massradius_conservative(self, hostname=None):
        """Plot Mass-Radius Diagram 
        Plot the mass radius diagram for the conservative habitable zone 

        Args:
            hostname (string)
        
        Returns:
            matplotlib.pyplot
        """
        exp = archive_explorer.ArchiveExplorer()
        if hostname:
            tab = exp.query_exo(hostname=hostname)
        else:
            tab = exp.query_exo()

        if tab.empty:
            print("No matching exoplanet data found.")
            return

        eval = hab_zone.HabZoneEvaluator()
        table = eval.conservative_habzone()
        cons_table = table[table['In Conservative Habitable Zone'] == True]
        cons_mass = cons_table['pl_msinie']
        cons_radii = cons_table['pl_rade']
        cons_temp = cons_table['st_teff'].values

        plt.scatter(cons_mass,cons_radii,c=cons_temp,cmap='inferno')
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('Minimum Mass (M$_{\oplus}$)')
        plt.ylabel('Planet Radius (R$_{\oplus}$)')
        plt.colorbar(label='Host Star $T_{eff}$')
        plt.title('Mass-Radius Relation for Planets in the Conservative HZ')
        plt.show()

        return
    
    def plot_massradius_optimistic(self, hostname=None):
        """Plot Mass-Radius Diagram 
        Plot the mass radius diagram for the optimistic habitable zone 

        Args:
            hostname (string)
        
        Returns:
            matplotlib.pyplot
        """
        exp = archive_explorer.ArchiveExplorer()
        if hostname:
            tab = exp.query_exo(hostname=hostname)
        else:
            tab = exp.query_exo()

        if tab.empty:
            print("No matching exoplanet data found.")
            return

        eval = hab_zone.HabZoneEvaluator()
        table2 = eval.optimistic_habzone()
        opt_table = table2[table2['In Optimistic Habitable Zone'] == True]
        opt_mass = opt_table['pl_msinie']
        opt_radii = opt_table['pl_rade']
        opt_temp = opt_table['st_teff'].values

        plt.scatter(opt_mass,opt_radii,c=opt_temp,cmap='inferno')
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('Minimum Mass (M$_{\oplus}$)')
        plt.ylabel('Planet Radius (R$_{\oplus}$)')
        plt.colorbar(label='Host Star $T_{eff}$')
        plt.title('Mass-Radius Relation for Planets in the Optimistic HZ')
        plt.show()

        return
    
